'use strict';

const logger = require('@firstfleet/fflogger');

/**
 * An object that holds error levels.
 * @type {{COMMON: string, FATAL: string, CRITICAL: string}}
 */
const errorLevels = {
    COMMON: 'common',
    CRITICAL: 'critical',
    FATAL: 'fatal'
};

/**
 * Creates a new AppError
 * @class
 */
class AppError {
    constructor(sender, program, message, severity, isOperational) {
        Error.call(this);
        Error.captureStackTrace(this);
        this.sender = sender;
        this.program = program;
        this.message = message;
        this.severity = severity;
        this.isOperational = isOperational;
    }
}

/**
 * Attach error prototypical inheritance
 */
AppError.prototype.__proto__ = Error.prototype;

const logError = (error) => {
    return new Promise((resolve, reject) => {
        try {
            logger.log('error', error);
            resolve('Error Logged')
        } catch(err) {
            reject(err)
        }
    });
};

/**
 * Handles application errors and calls the ffLogger (sends log to console if not
 * in production and also to papertrail). Collects params and stack trace.
 * @param sender - Machine sending the error
 * @param program - App sending the error
 * @param message - error description
 * @param severity - which errorLevel (defined in error handler)
 * @param isOperational - is it an operational error or programmer error
 * @returns {Promise<any>}
 */
const handleError = (sender, program, message, severity, isOperational) => {
    return new Promise((resolve, reject) => {
        const error = new AppError(sender, program, message, severity, isOperational);
        return logError(error)
            .then(error => resolve(error))
            .catch(caughtError => {
                console.error(caughtError);
                reject(caughtError)
            })
    });
};

/**
 * Takes in a custom error object and return a boolean. If the error
 * is an operation error it returns true, and returns false if it is a
 * Programmer error.
 * @param error
 * @returns {*}
 */
const isTrustedError = (error) => {
    return error.isOperational;
};

/**
 * Global handler for all uncaught exceptions
 * If an uncaughtException is not an Operational error,
 * gracefully restart the application.
 */
process.on('uncaughtException', (error) => {
    handleError('uncaughtException', error.message, errorLevels.CRITICAL, false)
        .then(_error => {
            if (!isTrustedError(_error)) {
                process.exit(1)
            }
        })
        .catch(caughtError => {
            console.error(error);
            process.exit(1)
        })

});

/**
 * Global handler for all uncaught promise rejections.
 * If an uncaught promise rejection is thrown this will handle the error
 * gracefully.
 */
process.on('unhandledRejection', error => {
    handleError(error);
});

/**
 * errorHandler module export.
 * @type {{AppError: AppError, isTrustedError: (function(*): *), handleError: (function(string, string, string, boolean): Promise<error | never>)}}
 */
module.exports = {
    AppError,
    handleError,
    isTrustedError,
    errorLevels
};
